import asyncio
import datetime
import json
import time
from base64 import b64decode

import aiohttp
import logzero


class PaClient:
    async def request(
        self, method, path, params: dict = None, _is_token_check_request: bool = False
    ):
        if not _is_token_check_request:
            self.logger.debug(f"{method} request {self._base + path}: {params}")
            await self.check_token()
        else:
            self.logger.debug(
                f"{method} request {self._base + path} for user {params['username']}"
            )
        if params is None:
            params = dict()

        if method == "GET":
            return await self._session.get(
                self._base + path, params=params, proxy=self._proxy
            )
        elif method == "POST":
            return await self._session.post(
                self._base + path, json=params, proxy=self._proxy
            )
        else:
            raise NotImplementedError

    async def check_token(self):
        if self._token_expire_time - time.time() < 30:  # less than 30s
            if not self.username or not self.password:  # logon using token only
                raise Exception(
                    "Cannot extend the session - login and password are required"
                )
            r = await self.request(
                "POST",
                "login/sign-in",
                {"username": self.username, "password": self.password},
                _is_token_check_request=True,
            )
            if r.status == 200:
                token = (await r.json()).get("data")
                if token:
                    self._session.headers["Authorization"] = f"Bearer {token}"
                    expire_time = json.loads(b64decode(token.split(".")[1] + "=="))[
                        "expires"
                    ]
                    self._token_expire_time = expire_time
                    return True
                raise Exception(
                    "Failed to get token. Something went wrong with the service"
                )
            if 400 <= r.status < 500:
                raise Exception(
                    f"Failed to get token. Check your login/password: {r.status}"
                )
            else:
                raise Exception(
                    f"Failed to get token. Probably the service is down: {r.status}"
                )
        return True

    def __init__(
        self,
        username: str = None,
        password: str = None,
        token: str = None,
        proxy: str = None,
        logger_level=logzero.WARNING,
        base_url="https://pyanalysis.ptsecurity.tech/api/v2/",
    ):
        self._proxy = None
        if proxy:
            if "://" not in proxy:
                raise Exception(
                    "Invalid proxy schema. Proper one starts with 'http://', 'socks4://', 'socks5://', etc. Attention: https proxies have limited support in aiohttp - https://docs.aiohttp.org/en/stable/client_advanced.html#proxy-support"
                )
            self._proxy = proxy

        self._session = aiohttp.ClientSession()
        self._base = base_url
        self.logger = logzero.setup_logger(level=logger_level)
        self.username = None
        self.password = None
        self._token_expire_time = -1

        if username and password:
            self.username = username
            self.password = password
        elif token:
            self._session.headers["Authorization"] = f"Bearer {token}"
            expire_time = json.loads(b64decode(token.split(".")[1] + "=="))["expires"]
            self._token_expire_time = expire_time
        else:
            raise Exception("You should pass (username, password) or token")

    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        await self._session.close()

    def __del__(self):
        try:
            loop = asyncio.get_event_loop()
            if loop.is_running():
                loop.create_task(self._session.close())
            else:
                loop.run_until_complete(self._session.close())
        except Exception:
            pass

    async def edit_password(self, new_password):
        resp = await self.request("GET", "user/edit", {"password": new_password})
        resp = await resp.json()
        return resp

    async def health(self):
        resp = await self.request("GET", "pyanalysis/health")
        resp = await resp.json()
        if resp.get("detail") == "Service is ok":
            return True
        else:
            raise Exception(resp.get("detail"))

    async def me(self):
        resp = await self.request("POST", "user/me")
        resp = await resp.json()
        return resp

    async def find_package(
        self,
        package_name: str,
        package_ver: str = None,
        wait_task_completion: bool = True,
    ):
        data = {"package_name": package_name}
        if package_ver:
            data["package_ver"] = package_ver
        resp = await self.request("GET", "pyanalysis/package", data)
        resp = await resp.json()
        if not wait_task_completion:
            return resp

        # for worst case scenarios - 3 tries
        tries = 3
        while True:
            if resp.get("status") != "Wait":
                return resp

            tries -= 1
            if tries == -1:
                raise Exception(
                    "Package analysis took too long, wait_task_completion=True"
                )

            status = "Pending"
            waits = 10
            while status not in ("Finished", "Crashed") and waits:
                self.logger.debug(
                    f"Status {package_name} {package_ver} {resp.get('task_id')} - {status}"
                )
                time.sleep(6)
                status = (await self.analysis_status(resp.get("task_id"))).get("detail")
                waits -= 1

            resp = await self.request("GET", "pyanalysis/package", data)
            resp = await resp.json()

    async def analysis_status(self, task_id: str):
        resp = await self.request("GET", "pyanalysis/task-tree", {"origin_id": task_id})
        """
        #  None of the tasks in Finished or Crashed state
        RUNNING = "Running"
        #  All the tasks are finished OR no tasks are found
        FINISHED = "Finished"
        #  One of the task in tree is in crashed state
        CRASHED = "Crashed"
        """
        return await resp.json()

    async def current_rules_version(self):
        resp = await self.request("GET", "pyanalysis/current_rules_version")
        resp = await resp.json()
        return resp

    async def package_badges(self, packages: list):
        resp = await self.request("POST", "pyanalysis/package_badges", packages)
        resp = await resp.json()
        return resp

    async def file_rescan(self, file_sha256: str):
        resp = await self.request(
            "GET", "pyanalysis/file_rescan", {"file_sha256": file_sha256}
        )
        resp = await resp.json()
        return resp

    async def kb_package(
        self,
        package_name: str,
        package_ver: str = None,
    ):
        data = {"package_name": package_name}
        if package_ver:
            data["package_ver"] = package_ver
        resp = await self.request("GET", "pyanalysis/knowledge/package", data)
        resp = await resp.json()
        return resp

    async def kb_package_versions(
        self,
        package_name: str,
    ):
        data = {"package_name": package_name}
        resp = await self.request("GET", "pyanalysis/knowledge/package_versions", data)
        resp = await resp.json()
        return resp

    async def kb_package_history(
        self,
        package_name: str,
    ):
        data = {"package_name": package_name}
        resp = await self.request("GET", "pyanalysis/knowledge/package_history", data)
        resp = await resp.json()
        return resp

    async def kb_package_developers_events(self, package_name: str):
        resp = await self.request(
            "GET",
            "pyanalysis/knowledge/package_developers_events",
            {"package_name": package_name},
        )
        resp = await resp.json()
        return resp

    async def kb_package_developers_statuses(self, package_name: str):
        resp = await self.request(
            "GET",
            "pyanalysis/knowledge/package_developers_statuses",
            {"package_name": package_name},
        )
        resp = await resp.json()
        return resp

    async def kb_developer_packages_events(self, pypi_user: str):
        resp = await self.request(
            "GET",
            "pyanalysis/knowledge/developer_packages_events",
            {"package_name": pypi_user},
        )
        resp = await resp.json()
        return resp

    async def kb_developer_packages_statuses(self, pypi_user: str):
        resp = await self.request(
            "GET",
            "pyanalysis/knowledge/developer_packages_statuses",
            {"package_name": pypi_user},
        )
        resp = await resp.json()
        return resp

    async def kb_file_meta(self, file_sha256: str):
        # simple sanity check
        if len(file_sha256) != 64 or not all(
            i in "0123456789abcdef" for i in file_sha256
        ):
            raise Exception("Invalid hash")

        resp = await self.request(
            "GET", "pyanalysis/knowledge/file_meta", {"file_sha256": file_sha256}
        )
        resp = await resp.json()
        return resp

    async def kb_package_files_meta(
        self, package_name: str, package_ver: str, page: int = 0
    ):
        resp = await self.request(
            "GET",
            "pyanalysis/knowledge/package_files_meta",
            {"package_name": package_name, "package_ver": package_ver, "page": page},
        )
        resp = await resp.json()
        return resp

    async def kb_file_bytes(self, file_sha256: str):
        # simple sanity check
        if len(file_sha256) != 64 or not all(
            i in "0123456789abcdef" for i in file_sha256
        ):
            raise Exception("Invalid hash")

        resp = await self.request(
            "GET", "pyanalysis/knowledge/file", {"file_sha256": file_sha256}
        )
        resp = await resp.json()
        return b64decode(resp["data"])

    async def kb_file_releases(self, file_sha256: str):
        # simple sanity check
        if len(file_sha256) != 64 or not all(
            i in "0123456789abcdef" for i in file_sha256
        ):
            raise Exception("Invalid hash")

        resp = await self.request(
            "GET", "pyanalysis/knowledge/file_releases", {"file_sha256": file_sha256}
        )
        resp = await resp.json()
        return resp

    async def feed_starjacking(self, limit: int = 5):
        resp = await self.request(
            "GET", "pyanalysis/feed/starjacking", {"limit": limit}
        )
        resp = await resp.json()
        return resp

    async def feed_typosquatting(self, limit: int = 5):
        resp = await self.request(
            "GET", "pyanalysis/feed/typosquatting", {"limit": limit}
        )
        resp = await resp.json()
        return resp

    async def feed_deleted(self, limit: int = 10):
        resp = await self.request("GET", "pyanalysis/feed/deleted", {"limit": limit})
        resp = await resp.json()
        return resp

    async def feed_osv(
        self,
        limit: int = 10,
        page: int = 1,
        since_date: datetime.datetime = None,
        since_report_id: int = None,
        only_withdrawn: bool = False,
        reversed: bool = False,
    ):
        resp = await self.request(
            "GET",
            "pyanalysis/feed/deleted",
            {
                "limit": limit,
                "page": page,
                "since_date": since_date,
                "since_report_id": since_report_id,
                "only_withdrawn": only_withdrawn,
                "reversed": reversed,
            },
        )
        resp = await resp.json()
        return resp



# sanity check
import os
import random
import string
from hashlib import sha256

import pytest

from pyanalysis.asyncio.client import PaClient

print(pytest.__version__)


@pytest.mark.asyncio
async def test_connectivity():
    pa_checker = PaClient(
        username=os.getenv("test_username"), password=os.getenv("test_password")
    )
    assert await pa_checker.health()


@pytest.mark.asyncio
async def test_connectivity_proxy():
    """ "
    pa_checker = PaClient(
        username=os.getenv("test_username"),
        password=os.getenv("test_password"),
        proxy=os.getenv("test_proxy"),
    )
    assert await pa_checker.health()
    """
    return True


@pytest.mark.asyncio
async def test_connectivity_with():
    async with PaClient(
        username=os.getenv("test_username"), password=os.getenv("test_password")
    ) as pa_checker:
        assert await pa_checker.health()


@pytest.mark.asyncio
async def test_get_user():
    pa_checker = PaClient(
        username=os.getenv("test_username"), password=os.getenv("test_password")
    )
    me = await pa_checker.me()
    assert me.get("username")
    assert me.get("email")
    assert me.get("disabled") in (True, False)
    assert me.get("disabled") is False  # heh, this check we definitely need)
    assert (
        me.get("password") == "<Hidden>"
    )  # no, you cannot use this password, it's just to prevent leak
    plan = me.get("plan")
    assert plan
    assert plan.get("month_plan")
    assert plan.get("used", None) is not None


@pytest.mark.asyncio
async def test_use_token():
    pa_checker1 = PaClient(
        username=os.getenv("test_username"), password=os.getenv("test_password")
    )
    # Need to mention: token obtaining in async version will occur after first api request
    # So we need to do something, e.g. check us
    await pa_checker1.me()
    pa_checker1_token = pa_checker1._session.headers["Authorization"].split(
        " ", maxsplit=1
    )[1]
    pa_checker2 = PaClient(token=pa_checker1_token)
    me = await pa_checker2.me()
    assert me.get("username")


@pytest.mark.asyncio
async def test_wrong_credentials():
    try:
        PaClient(
            username="".join(random.choices(string.ascii_letters, k=16)),
            password="do not use this password :)",
        )
        return False
    except Exception:
        return True


@pytest.mark.asyncio
async def test_known_package():
    pa_checker = PaClient(
        username=os.getenv("test_username"), password=os.getenv("test_password")
    )
    resp = await pa_checker.find_package("requests")
    assert resp.get("status")
    assert resp.get("why")
    assert resp.get("weights_by_versions")
    assert resp.get("verdicts_by_versions")
    assert resp.get("files")
    assert resp.get("rules_score")


@pytest.mark.asyncio
async def test_known_package_with_version():
    pa_checker = PaClient(
        username=os.getenv("test_username"), password=os.getenv("test_password")
    )
    resp = await pa_checker.find_package("requests", "2.11.1")
    assert resp.get("status")
    assert resp.get("why")
    assert resp.get("weights_by_versions")
    assert len(resp.get("weights_by_versions")) == 1
    assert resp.get("verdicts_by_versions")
    assert len(resp.get("verdicts_by_versions")) == 1
    assert resp.get("files")
    assert resp.get("rules_score")


@pytest.mark.asyncio
async def test_non_existent_package():
    pa_checker = PaClient(
        username=os.getenv("test_username"), password=os.getenv("test_password")
    )
    resp = await pa_checker.find_package(
        "".join(random.choices(string.ascii_letters, k=16))
    )
    assert resp.get("status") == "Not found"
    assert "404 Status code on pypi.org" in resp.get("why")


@pytest.mark.asyncio
async def test_non_existent_version():
    pa_checker = PaClient(
        username=os.getenv("test_username"), password=os.getenv("test_password")
    )
    resp = await pa_checker.find_package("requests", "0.0.0.0.0")
    assert resp.get("status") == "Not found"
    assert "This package with your package_ver not found" in resp.get("why")
    print(resp)


@pytest.mark.asyncio
async def test_knowledge_base_get_file():
    pa_checker = PaClient(
        username=os.getenv("test_username"), password=os.getenv("test_password")
    )
    resp = await pa_checker.kb_package("requests", "2.11.1")
    assert resp.get("data")
    package_data = resp.get("data")[0]  # if it is only one package with this version
    file_resp = await pa_checker.kb_file_meta(file_sha256=package_data.get("py")[0])
    file_hash = file_resp.get("data").get("_id")
    assert file_hash == package_data.get("py")[0]

    file_data = await pa_checker.kb_file_bytes(file_sha256=file_hash)
    assert file_hash == sha256(file_data).hexdigest()


@pytest.mark.asyncio
async def test_knowledge_base_non_existent_package():
    pa_checker = PaClient(
        username=os.getenv("test_username"), password=os.getenv("test_password")
    )
    resp = await pa_checker.kb_package(
        "".join(random.choices(string.ascii_letters, k=16))
    )
    assert resp.get("status") == "Not found"
    assert "404 Status code on pypi.org" in resp.get("why")


@pytest.mark.asyncio
async def test_knowledge_base_non_existent_version():
    pa_checker = PaClient(
        username=os.getenv("test_username"), password=os.getenv("test_password")
    )
    resp = await pa_checker.kb_package("requests", "0.0.0.0.0")
    assert resp.get("status") == "Not found"
    assert "This package with your package_ver not found" in resp.get("why")


@pytest.mark.asyncio
async def test_knowledge_base_get_packave_version_list():
    pa_checker = PaClient(
        username=os.getenv("test_username"), password=os.getenv("test_password")
    )
    resp = await pa_checker.kb_package_versions("requests")
    assert resp.get("data")
    one_data = resp.get("data")[0]
    assert one_data.get("project") == "requests"
    assert one_data.get("project_ver")
    assert one_data.get("download_url")
    assert one_data.get("creation_date")


@pytest.mark.asyncio
async def test_knowledge_base_get_packave_history():
    pa_checker = PaClient(
        username=os.getenv("test_username"), password=os.getenv("test_password")
    )
    resp = await pa_checker.kb_package_history("requests")
    assert resp.get("data")
    one_data = resp.get("data")[0]
    assert one_data.get("project_name") == "requests"
    assert one_data.get("description")


@pytest.mark.asyncio
async def test_knowledge_base_get_file_releases():
    pa_checker = PaClient(
        username=os.getenv("test_username"), password=os.getenv("test_password")
    )
    resp = await pa_checker.kb_package("requests", "2.11.1")
    assert resp.get("data")
    package_data = resp.get("data")[0]  # if it is only one package with this version

    file_resp = await pa_checker.kb_file_releases(file_sha256=package_data.get("py")[0])
    #  raise Exception( f"{package_data}\n\n{file_resp}")
    assert any(
        package_data.get("_id") == fresp.get("_id") for fresp in file_resp.get("data")
    )


@pytest.mark.asyncio
async def test_task_status():
    pa_checker = PaClient(
        username=os.getenv("test_username"), password=os.getenv("test_password")
    )
    resp = await pa_checker.analysis_status("00000000-0000-0000-0000-000000000000")
    assert resp.get("detail") == "Finished"


async def test_osv_feed():
    pa_checker = PaClient(
        username=os.getenv("test_username"), password=os.getenv("test_password")
    )
    resp = await pa_checker.feed_osv()
    assert resp.get("data")
